use strict;
use warnings;

return [
  {
    'author' => [
      'Dominique Dumont'
    ],
    'copyright' => [
      '2010,2011 Dominique Dumont'
    ],
    'element' => [
      'rock',
      {
        'type' => 'leaf',
        'value_type' => 'boolean'
      },
      'joliet',
      {
        'type' => 'leaf',
        'value_type' => 'boolean'
      }
    ],
    'include' => [
      'Fstab::CommonOptions'
    ],
    'license' => 'LGPL2',
    'name' => 'Fstab::Iso9660_Opt'
  }
]
;

